/*
 * Copyright (c) 2014-2018, Sonos, Inc.
 *
 * SPDX-License-Identifier:	GPL-2.0
 *
 * sonos_fw_whitelist.c.inc: implementation of check_sonos_firmware_whitelist
 *
 * !!NOTE!! this file is meant to be pulled in with #include after the
 * including project has defined some macros:
 *      int SFW_GETCPUID(uint8_t *buf, size_t len)
 *      int SFW_GETSERIAL(uint8_t *buf, size_t len)
 *      SFW_PRINT(fmt, fmtArgs...)
 *      SFW_PLVL_INFO
 *      SFW_PLVL_EMERG
 *      SFW_BE32_TO_CPU(x)
 *
 * Note that SWF_GETSERIAL should return true if it can get a known
 * good (cryptographically secure) value for serial; false otherwise.
 * If it fails then any non-empty whitelist causes us to not boot.
 *
 * Callers needing to validate whitelists in other images should also define:
 *      SFW_INCLUDE_FIND_AND_CHECK_SONOS_WHITELIST
 *
 * Trying to compile this as a .o and link it in multiple places doesn't work
 * well in situations like the u-boot build or the kernel build (hence the
 * weird file inclusion tactic used here).
 */

static void
initialPrintHelper(uint8_t type, const uint8_t* id)
{
    const char* typeName;

    if (type == SONOS_FWW_TYPE_SERIAL) {
        typeName = "serial";
        SFW_PRINT(SFW_PLVL_INFO
                  "checking %s whitelist (my %s is "
                  SONOS_FWW_FMT_SERIAL ")...\n",
                  typeName, typeName, SONOS_FWW_FMTARG_SERIAL(id));
    }
    else {
        typeName = "cpuid";
        SFW_PRINT(SFW_PLVL_INFO
                  "checking %s whitelist (my %s is "
                  SONOS_FWW_OWN_FMT_CPUID ")...\n",
                  typeName, typeName, SONOS_FWW_OWN_FMTARG_CPUID(id));
    }
    (void)typeName;
}

static void
foreachPrintHelper(uint8_t type, const uint8_t* id)
{
    if (type == SONOS_FWW_TYPE_SERIAL) {
        SFW_PRINT(SFW_PLVL_INFO
                  "checking serial "SONOS_FWW_FMT_SERIAL"\n",
                  SONOS_FWW_FMTARG_SERIAL(id));
    }
    else {
        SFW_PRINT(SFW_PLVL_INFO
                  "checking cpuid "SONOS_FWW_OWN_FMT_CPUID"\n",
                  SONOS_FWW_OWN_FMTARG_CPUID(id));
    }
}

int
check_sonos_firmware_whitelist_ex(const SonosFirmwareWhitelistHeader* hdr,
                                  const uint8_t *entries,
                                  size_t maxEntriesSize)
{
    uint32_t numEntries = SFW_BE32_TO_CPU(hdr->numEntries);
    uint8_t type = hdr->whitelistType;
    int foundWhitelistEntry = 0;
    const uint8_t* myId = NULL;
    const uint8_t* listId = entries;
    size_t listIdLen;
    const char* typeName = NULL;
    uint32_t i;
    SonosOwnCpuid_t myCpuid;
    SonosSerial_t mySerial;
    int result = 0;

    if (type != SONOS_FWW_TYPE_CPUID && type != SONOS_FWW_TYPE_SERIAL && type != SONOS_FWW_TYPE_CPUID16) {
        SFW_PRINT(SFW_PLVL_EMERG
                  "bad Sonos firmware whitelist type: %d\n", (int)type);
        return result;
    }
    if (numEntries > SONOS_FWW_MAX_ENTRIES) {
        SFW_PRINT(SFW_PLVL_EMERG
                  "Firmware whitelist too large (%lu)\n", (unsigned long)numEntries);
        return result;
    }

    if (type == SONOS_FWW_TYPE_SERIAL) {
        typeName = "serial";
        if (SFW_GETSERIAL(mySerial.value, sizeof(mySerial.value))) {
            myId = mySerial.value;
        }
        listIdLen = sizeof(mySerial);
    }
    else {
        typeName = "cpuid";
        if (SFW_GETCPUID(myCpuid.value, sizeof(myCpuid.value))) {
            myId = myCpuid.value;
        }
        listIdLen = sizeof(myCpuid);
    }

    if (numEntries * listIdLen > maxEntriesSize) {
        SFW_PRINT(SFW_PLVL_EMERG "Firmware whitelist invalid size\n");
        return result;
    }

    if (myId) {
        initialPrintHelper(type, myId);

        if (numEntries != 0) {
            if (type != SONOS_FWW_TYPE_SERIAL &&
                type != SONOS_FWW_OWN_TYPE_CPUID) {
                SFW_PRINT(SFW_PLVL_EMERG
                          "bad Sonos firmware whitelist type: %d\n", (int)type);
                return result;
            }

            for (i = 0; i < numEntries; i++, listId += listIdLen) {
                if (i < 10) {
                    foreachPrintHelper(type, listId);
                }
                else if (i == 10) {
                    SFW_PRINT(SFW_PLVL_INFO
                              "... checking too many to list ...\n");
                }
                if (memcmp(myId, listId, listIdLen) == 0) {
                    foundWhitelistEntry = 1;
                    break;
                }
            }
            if (!foundWhitelistEntry) {
                SFW_PRINT(SFW_PLVL_EMERG
                          "failed %s whitelist check\n", typeName);
                return 0;
            }
        }
    }
    else {
        SFW_PRINT(SFW_PLVL_EMERG
                  "checking %s whitelist against unknown %s...\n",
                  typeName, typeName);

        if (numEntries != 0) {
            SFW_PRINT(SFW_PLVL_EMERG
                      "non-empty (%d entries) Sonos %s whitelist; dying...\n",
                      (int)numEntries, typeName);
            return 0;
        }
    }

    SFW_PRINT(SFW_PLVL_INFO "whitelist check completed\n");
    (void)typeName;
    return 1;
}


static inline int
check_sonos_firmware_whitelist(const SonosFirmwareWhitelistHeader* hdr, size_t entriesSize)
{
    return check_sonos_firmware_whitelist_ex(hdr, (const uint8_t *)&hdr[1], entriesSize);
}

#ifdef SFW_INCLUDE_FIND_AND_CHECK_SONOS_WHITELIST

int find_and_check_sonos_whitelist(const uint8_t* buf, size_t bufLen,
                                   int isKernel, int isPermissive)
{
    static const uint8_t magicUBoot[] = { SONOS_FWW_MAGIC_INIT_INVERTED };
    static const uint8_t magicKernel[] = { SONOS_FWW_MAGIC_INIT_UNCOMPRESSED_COPY_INVERTED };
    uint8_t magic[SONOS_FWW_MAGIC_LEN];
    const uint8_t* pMagic;
    const uint8_t* p;
    const uint8_t* pEnd;
    const uint8_t* pEntries = NULL;
    size_t i;
    SonosFirmwareWhitelistHeader whitelistHdr;

    memset(&whitelistHdr, 0, sizeof(whitelistHdr));

    if (bufLen < sizeof(SonosFirmwareWhitelistHeader)) {
        return 0;
    }

    pMagic = isKernel ? magicKernel : magicUBoot;
    for (i = 0; i < sizeof(magic); i++) {
        magic[i] = pMagic[i] ^ 0xff;
    }

    pEnd = &buf[bufLen - sizeof(SonosFirmwareWhitelistHeader)];

    for (p = buf; p <= pEnd; p++) {
        if (memcmp(p, magic, sizeof(magic)) == 0) {
            if (pEntries) {
                SFW_PRINT(SFW_PLVL_EMERG "Duplicate whitelist\n");
                return 0;
            }
            memcpy(&whitelistHdr, p, sizeof(whitelistHdr));
            pEntries = p + sizeof(SonosFirmwareWhitelistHeader);
        }
    }

    if (!pEntries) {
        if (isPermissive) {
            return 1;
        }
        return 0;
    }

    return check_sonos_firmware_whitelist_ex(&whitelistHdr, pEntries,
                                             &buf[bufLen] - pEntries);
}

#endif

